/*
 * Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file send.h Sender-related defines and structures. */

#define SENDER_SUBCOMMANDS \
	SENDER_SUBCOMMAND(add) /**< Add a target (udp only). */ \
	SENDER_SUBCOMMAND(delete) /**< Delete a target (udp only). */ \
	SENDER_SUBCOMMAND(allow) /**< Allow connections from given IP address(es). */ \
	SENDER_SUBCOMMAND(deny) /**< Deny connections from given IP address(es). */ \
	SENDER_SUBCOMMAND(on) /**< Activate the sender. */ \
	SENDER_SUBCOMMAND(off) /**< Deactivate the sender. */ \

#define SENDER_SUBCOMMAND(_name) SENDER_ ## _name,
enum sender_subcommand {
	SENDER_SUBCOMMANDS
	NUM_SENDER_CMDS /**< Used as array size in struct \ref sender. */
};
#undef SENDER_SUBCOMMAND
#define SENDER_SUBCOMMAND(_name) #_name,

/**
 * Describes one supported sender of para_server.
 *
 * \sa \ref http_send.c \ref udp_send.c, \ref dccp_send.c.
 */
struct sender {
	/** The name of the sender. */
	const char *name;
	/**
	 * The init function of this sender.
	 *
	 * It must fill in all function pointers of \a s as well as the \a
	 * client_cmds array, see below. It should also do all necessary
	 * preparations to init this sending facility, for example it could
	 * open a tcp port.
	 */
	void (*init)(struct sender *s);
	/**
	 * Return the help text of this sender.
	 *
	 * The result must be dynamically allocated and is freed by the caller.
	 */
	char* (*help)(void);
	/**
	 * Return current status info about this sender.
	 *
	 * The result must be dynamically allocated and is freed by the caller.
	 */
	char* (*status)(void);
	/**
	 * The send-hook.
	 *
	 * It gets called whenever para_server is playing and the current
	 * audio format handler indicates that another chunk of data should
	 * be sent now. The two parameters \a current_chunk and \a chunks_sent
	 * only differ if the stream was repositioned by the \a ff or \a jmp
	 * command. Of course, \a buf is a pointer to the chunk of data which
	 * should be sent, and \a len is the length of this buffer.
	 */
	void (*send)(long unsigned current_chunk, long unsigned chunks_sent,
		const char *buf, size_t len, const char *header_buf,
		size_t header_len);
	/**
	 * Add file descriptors to fd_sets.
	 *
	 * The pre_select function of each supported sender is called just before
	 * para_server enters its main select loop. Each sender may add its own
	 * file descriptors to the \a rfds or the \a wfds set.
	 *
	 * If a file descriptor was added, \a max_fileno must be increased by
	 * this function, if necessary.
	 *
	 * \sa select(2).
	 */
	void (*pre_select)(int *max_fileno, fd_set *rfds, fd_set *wfds);
	/**
	 * Handle the file descriptors which are ready for I/O.
	 *
	 * If the pre_select hook added one ore more file descriptors to the
	 * read or write set, this is the hook to check the result and do any
	 * I/O on those descriptors which are ready for reading/writing.
	 */
	void (*post_select)(fd_set *rfds, fd_set *wfds);
	/**
	 * Terminate all connected clients.
	 *
	 * This is called e.g. if the stop command was executed. It should make
	 * the clients aware of the end-of-file condition.
	 */
	void (*shutdown_clients)(void);
	/**
	 * Array of function pointers for the sender subcommands.
	 *
	 * Each sender may implement any subset of the sender commands by
	 * filling in the appropriate function pointer in the array. A \p NULL
	 * pointer means this command is not implemented by this sender.
	 */
	int (*client_cmds[NUM_SENDER_CMDS])(struct sender_command_data*);
	/**
	 * Resolve target-specific URL string
	 *
	 * This method must be defined if the sender supports the add/delete
	 * subcommands. It interprets a string specifying a target URL in a
	 * sender-specific fashion (e.g. embedded FEC string). It can also
	 * fill in sender-specific defaults if necessary.
	 */
	int (*resolve_target)(const char *, struct sender_command_data *);
};

/** Describes one client, connected to a paraslash sender. */
struct sender_client {
	/** The file descriptor of the client. */
	int fd;
	/** The socket "name" of the client. */
	char *name;
	/** The position of this client in the client list. */
	struct list_head node;
	/** Non-zero if audio file header has been sent. */
	int header_sent;
	/** The list of pending chunks for this client. */
	struct chunk_queue *cq;
	/** Data specific to the particular sender. */
	void *private_data;
};

/**
 * Each paraslash sender may register arbitrary many clients to the virtual
 * streaming system, possibly with varying fec parameters. In order to do so,
 * it must allocate a \a fec_client_parms structure and pass it to \ref
 * vss_add_fec_client.
 *
 * Clients are automatically removed from that list by the vss if an error
 * occurs, or if the sender requests deletion of a client by calling \ref
 * vss_del_fec_client().
 */

/** FEC parameters requested by FEC clients. */
struct fec_client_parms {
	/** Number of data slices plus redundant slices. */
	uint8_t slices_per_group;
	/** Number of slices minus number of redundant slices. */
	uint8_t data_slices_per_group;
	/** Whether the header must be sent periodically. */
	bool need_periodic_header;
	/**
	 * Transport-layer initialisation for FEC support.
	 *
	 * This optional function serves (a) to make the transport layer
	 * ready to send FEC slices and (b) to determine the Maximum
	 * Packet Size (MPS) supported by the connection. The MPS value
	 * determines the largest payload size. This value is used to
	 * send FEC slices that are not larger than the path MTU, to avoid
	 * fragmentation and to maximize packet utilization. The user can
	 * alternatively specify a slice size of up to this value.
	 */
	int (*init_fec)(struct sender_client *sc);
	/** Push out FEC-encoded packets */
	void (*send_fec)(struct sender_client *sc, char *buf, size_t len);
};

/** Describes the current status of one paraslash sender. */
struct sender_status {
	/** The file descriptor of the socket this sender is listening on. */
	int listen_fd;
	/** The TCP/DCCP port used by this sender. */
	int port;
	/** The current number of simultaneous connections. */
	int num_clients;
	/** The maximal number of simultaneous connections. */
	int max_clients;
	/** Whether the access control list is a whitelist. */
	int default_deny;
	/** The whitelist/blacklist. */
	struct list_head acl;
	/** The list of connected clients. */
	struct list_head client_list;
};

void shutdown_client(struct sender_client *sc, struct sender_status *ss);
void shutdown_clients(struct sender_status *ss);
void init_sender_status(struct sender_status *ss,
		const struct lls_opt_result *acl_opt_result, int port,
		int max_clients, int default_deny);
char *generic_sender_status(struct sender_status *ss, const char *name);
void generic_com_allow(struct sender_command_data *scd,
		struct sender_status *ss);
void generic_com_deny(struct sender_command_data *scd,
		struct sender_status *ss);
int generic_com_on(struct sender_status *ss, unsigned protocol);
void generic_com_off(struct sender_status *ss);
char *generic_sender_help(void);
struct sender_client *accept_sender_client(struct sender_status *ss, fd_set *rfds);
int send_queued_chunks(int fd, struct chunk_queue *cq);
int parse_fec_url(const char *arg, struct sender_command_data *scd);
